/*

SynoDL - CLI for Synology's DownloadStation
Copyright (C) 2015 - 2023, 2025  Stefan Ott

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, version 3.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

use std::error;
use std::io::{self, ErrorKind};
#[cfg(feature = "custom_ca")]
use ureq::tls::RootCerts;
use ureq::tls::{Certificate, TlsConfig};

#[cfg(feature = "custom_ca")]
use {
    rustls_pemfile::{read_one, Item},
    std::{fs::File, io::BufReader, iter},
};

#[cfg(test)]
use mockall::automock;

use crate::Config;

fn tls_default() -> Result<TlsConfig, Box<dyn error::Error>> {
    Ok(TlsConfig::builder().build())
}

#[cfg(feature = "custom_ca")]
fn get_cert(path: &String) -> Option<Certificate<'static>> {
    let Ok(mut f) = File::open(&path) else {
        return None;
    };

    let mut crt = BufReader::new(&mut f);
    for item in iter::from_fn(|| read_one(&mut crt).transpose()) {
        match item.unwrap() {
            Item::X509Certificate(cert) => {
                let der: &[u8] = &cert;
                let c = Certificate::from_der(der);
                return Some(c.to_owned());
            }
            _ => println!("unhandled item"),
        }
    }

    None
}

#[cfg(feature = "custom_ca")]
fn tls_custom_cert(path: &String) -> Result<TlsConfig, Box<dyn error::Error>> {
    match get_cert(path) {
        Some(c) => {
            let certs = RootCerts::new_with_certs(&[c]);
            Ok(TlsConfig::builder().root_certs(certs).build())
        }
        None => Err(Box::new(io::Error::new(
            ErrorKind::Other,
            "Failed to load CA certificate",
        ))),
    }
}

#[cfg(not(feature = "custom_ca"))]
fn tls_custom_cert(_path: &String) -> Result<TlsConfig, Box<dyn error::Error>> {
    Err(Box::new(io::Error::new(
        ErrorKind::Other,
        "Please enable 'custom_ca' feature to specify your own CA certificate",
    )))
}

fn tls_ignore_cert() -> Result<TlsConfig, Box<dyn error::Error>> {
    let tc = TlsConfig::builder().disable_verification(true).build();
    Ok(tc)
}

#[cfg_attr(test, automock)]
trait TLSCertificateBehaviour {
    fn default_certs(&self) -> Result<TlsConfig, Box<dyn error::Error>>;
    fn custom_cert(&self, path: &String) -> Result<TlsConfig, Box<dyn error::Error>>;
    fn ignore_cert(&self) -> Result<TlsConfig, Box<dyn error::Error>>;
}

struct MyCertificateBehaviour {}

impl TLSCertificateBehaviour for MyCertificateBehaviour {
    fn default_certs(&self) -> Result<TlsConfig, Box<dyn error::Error>> {
        tls_default()
    }
    fn custom_cert(&self, path: &String) -> Result<TlsConfig, Box<dyn error::Error>> {
        tls_custom_cert(path)
    }
    fn ignore_cert(&self) -> Result<TlsConfig, Box<dyn error::Error>> {
        tls_ignore_cert()
    }
}

fn load_tls_config(
    cacert: &Option<String>,
    options: &impl TLSCertificateBehaviour,
) -> Result<TlsConfig, Box<dyn error::Error>> {
    match cacert {
        Some(f) => {
            if f == "ignore" {
                options.ignore_cert()
            } else {
                options.custom_cert(&f)
            }
        }
        None => options.default_certs(),
    }
}

pub fn get_tls_config(cfg: &Config) -> Result<TlsConfig, Box<dyn error::Error>> {
    load_tls_config(&cfg.cacert, &MyCertificateBehaviour {})
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_default_tls_config() {
        let mut mock = MockTLSCertificateBehaviour::new();
        let client = TlsConfig::builder().build();
        let client2 = client.clone();
        let client3 = client.clone();

        mock.expect_default_certs()
            .times(1)
            .return_once(|| Ok(client));
        mock.expect_custom_cert()
            .times(0)
            .return_once(|_| Ok(client2));
        mock.expect_ignore_cert()
            .times(0)
            .return_once(|| Ok(client3));

        let _ = load_tls_config(&None, &mock);
    }

    #[test]
    fn test_tls_config_custom_certificate() {
        let mut mock = MockTLSCertificateBehaviour::new();
        let client = TlsConfig::builder().build();
        let client2 = client.clone();
        let client3 = client.clone();

        mock.expect_default_certs()
            .times(0)
            .return_once(|| Ok(client));
        mock.expect_custom_cert()
            .times(1)
            .return_once(|_| Ok(client2));
        mock.expect_ignore_cert()
            .times(0)
            .return_once(|| Ok(client3));

        let _ = load_tls_config(&Some(String::from("/tmp/test.pem")), &mock);
    }

    #[test]
    fn test_tls_config_ignore_certificate() {
        let mut mock = MockTLSCertificateBehaviour::new();
        let client = TlsConfig::builder().build();
        let client2 = client.clone();
        let client3 = client.clone();

        mock.expect_default_certs()
            .times(0)
            .return_once(|| Ok(client));
        mock.expect_custom_cert()
            .times(0)
            .return_once(|_| Ok(client2));
        mock.expect_ignore_cert()
            .times(1)
            .return_once(|| Ok(client3));

        let _ = load_tls_config(&Some(String::from("ignore")), &mock);
    }
}
